VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
  Persistable = 0  'NotPersistable
  DataBindingBehavior = 0  'vbNone
  DataSourceBehavior  = 0  'vbNone
  MTSTransactionMode  = 0  'NotAnMTSObject
END
Attribute VB_Name = "ExpressionNode"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = True
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = False
Option Explicit

Implements IStatement

Public Enum NODE_TYPE
    NODE_NOT_DEFINED = 0
    NODE_OP_ASSIGN
    NODE_OP_ADD
    NODE_OP_SUB
    NODE_OP_MUL
    NODE_OP_DIV
    NODE_OP_NEG
    NODE_OP_CMP_EQU
    NODE_OP_CMP_NEQU
    NODE_OP_CMP_GRT
    NODE_OP_CMP_GRTEQU
    NODE_OP_CMP_SML
    NODE_OP_CMP_SMLEQU
    NODE_OP_LGC_AND
    NODE_OP_LGC_OR
    NODE_OP_BAND
    NODE_OP_BOR
    NODE_OP_BXOR
    NODE_OP_MOD
    NODE_OP_CALL
    NODE_OP_DEREF_ARR
    NODE_VAL_FALSE
    NODE_VAL_TRUE
    NODE_VAL_VALUE
    NODE_VAL_VARIABLE
    NODE_CONV_BYTE_TO_SHORT
    NODE_CONV_BYTE_TO_INT
    NODE_CONV_BYTE_TO_FLOAT
    NODE_CONV_SHORT_TO_INT
    NODE_CONV_SHORT_TO_FLOAT
    NODE_CONV_INT_TO_FLOAT
End Enum

Public Enum DATA_TYPE
    TYPE_UNKNOWN = -1
    TYPE_NONE = 0
    TYPE_INT = &H1
    TYPE_SHORT = &H2
    TYPE_BYTE = &H4
    TYPE_FLOAT = &H8
    TYPE_BOOL = &H10
    TYPE_ARRAY = &H1000
End Enum

Private m_clsParent         As ExpressionNode
Private m_clsSubNodes()     As ExpressionNode
Private m_lngSubNodeCount   As Long

Private m_udeNdType         As NODE_TYPE
Private m_udeDtType         As DATA_TYPE
Private m_clsArrIndex       As ExpressionNode

Private m_strIdentifier     As String
Private m_blnIsRef          As Boolean
Private m_varValue          As Variant

Public Property Get IsReference() As Boolean
    IsReference = m_blnIsRef
End Property

Public Property Let IsReference(ByVal blnVal As Boolean)
    m_blnIsRef = blnVal
End Property

Public Property Get ArrayIndex() As ExpressionNode
    Set ArrayIndex = m_clsArrIndex
End Property

Public Property Set ArrayIndex(ByVal cls As ExpressionNode)
    Set m_clsArrIndex = cls
End Property

Public Property Get Identifier() As String
    Identifier = m_strIdentifier
End Property

Public Property Let Identifier(ByVal strVal As String)
    m_strIdentifier = strVal
End Property

Public Property Get Value() As Variant
    Value = m_varValue
End Property

Public Property Let Value(ByVal varVal As Variant)
    m_varValue = varVal
End Property

Public Property Get Parent() As ExpressionNode
    Set Parent = m_clsParent
End Property

Public Property Set Parent(ByVal cls As ExpressionNode)
    Set m_clsParent = cls
End Property

Public Property Get SubNodeCount() As Long
    SubNodeCount = m_lngSubNodeCount
End Property

Public Property Get SubNode(ByVal i As Long) As ExpressionNode
    Set SubNode = m_clsSubNodes(i)
End Property

Public Property Set SubNode(ByVal i As Long, ByVal cls As ExpressionNode)
    Set m_clsSubNodes(i) = cls
End Property

Public Sub RemSubNode(ByVal nd As ExpressionNode)
    Dim i As Long, j As Long
    For i = 0 To m_lngSubNodeCount - 1
        If m_clsSubNodes(i) Is nd Then
            Set m_clsSubNodes(i) = Nothing
            For j = i To m_lngSubNodeCount - 2
                Set m_clsSubNodes(j) = m_clsSubNodes(j + 1)
            Next
            If m_lngSubNodeCount = 1 Then
                Erase m_clsSubNodes
            Else
                ReDim Preserve m_clsSubNodes(m_lngSubNodeCount - 2)
            End If
            m_lngSubNodeCount = m_lngSubNodeCount - 1
            Exit Sub
        End If
    Next
End Sub

Public Sub AddSubNode(ByVal nd As ExpressionNode)
    ReDim Preserve m_clsSubNodes(m_lngSubNodeCount) As ExpressionNode
    Set m_clsSubNodes(m_lngSubNodeCount) = nd
    m_lngSubNodeCount = m_lngSubNodeCount + 1
End Sub

Public Property Get NodeType() As NODE_TYPE
    NodeType = m_udeNdType
End Property

Public Property Let NodeType(ByVal ude As NODE_TYPE)
    m_udeNdType = ude
End Property

Public Property Get DataType() As DATA_TYPE
    DataType = m_udeDtType
End Property

Public Property Let DataType(ByVal ude As DATA_TYPE)
    m_udeDtType = ude
End Property

Private Sub IStatement_DeclareVars(ByVal ctx As TACContext)
    '
End Sub

Private Sub IStatement_DestroyVars(ByVal ctx As TACContext)
    '
End Sub

Private Function IStatement_GenerateTAC(ByVal ctx As TACContext) As TACVar
    Dim clsVarAsgn As TACVar
    Dim clsVarLeft As TACVar
    Dim clsVarRight As TACVar
    Dim clsVarSize As TACVar
    Dim lngVarSize As Long
    
    Select Case NodeType
        ' Variable kann Array oder auch Referenz sein!
        Case NODE_VAL_VARIABLE
            Set clsVarAsgn = CreateTACVar(TAC_VARIABLE, Identifier, DType:=DataType)
            
            If Not ArrayIndex Is Nothing Then
                If IsReference Then
                    ' Referenziertes Array dereferenzieren
                    clsVarAsgn.DeRef = True
                    Set clsVarLeft = CreateTempTACVar(ctx, DataType)
                    ctx.AddCode CreateTAC(TAC_CMD_ASSIGN, NODE_OP_ASSIGN, , clsVarLeft, clsVarAsgn)
                    Set clsVarAsgn = clsVarLeft
                End If
                
                ' Index kompilieren
                Set clsVarLeft = ToIStatement(ArrayIndex).GenerateTAC(ctx)
                If clsVarLeft.VarType = TAC_VARIABLE Or clsVarLeft.VarType = TAC_VALUE Then
                    ' Inhalt von Variable darf hier nicht verndert werden => kopieren
                    Dim clsVarCpy As TACVar
                    Set clsVarCpy = CreateTempTACVar(ctx, clsVarLeft.DataType)
                    ctx.AddCode CreateTAC(TAC_CMD_ASSIGN, NODE_OP_ASSIGN, , clsVarCpy, clsVarLeft)
                    Set clsVarLeft = clsVarCpy
                End If
                
                ' Index = Index * Len(Type)
                lngVarSize = ctx.ParserGlobal.VarBytes(DataType)
                Set clsVarSize = CreateTACVar(TAC_VALUE, , lngVarSize, TYPE_INT)
                ctx.AddCode CreateTAC(TAC_CMD_ASSIGN, NODE_OP_MUL, , clsVarLeft, clsVarLeft, clsVarSize)
                
                ' Index zu Array Pointer addieren
                ctx.AddCode CreateTAC(TAC_CMD_ASSIGN, NODE_OP_ADD, , clsVarLeft, clsVarAsgn, clsVarLeft)
                Set clsVarAsgn = CopyTACVar(clsVarLeft)
                
                ' Als zu dereferenzierend markieren
                clsVarAsgn.DeRef = True
            Else
                clsVarAsgn.DeRef = IsReference
            End If
        
        ' Wert als solchen einfach bergeben
        Case NODE_VAL_FALSE, NODE_VAL_TRUE, NODE_VAL_VALUE
            Set clsVarAsgn = CreateTACVar(TAC_VALUE, , Value, DataType)
        
        ' Call wird nicht als Teil eines Ausdrucks gesehen
        Case NODE_OP_CALL
            Dim i As Long
            Dim clsFnc As ParserFunction
            
            Set clsFnc = ctx.ParserGlobal.Functions(Identifier)
            
            For i = 0 To SubNodeCount - 1
                Set clsVarLeft = ToIStatement(SubNode(i)).GenerateTAC(ctx)
                clsVarLeft.DoRef = clsFnc.ParamIsRef(i)
                ctx.AddCode CreateTAC(TAC_CMD_PUSH, , , clsVarLeft)
            Next
            
            ctx.AddCode CreateTAC(TAC_CMD_CALL, , "fnc_" & Identifier)
            If clsFnc.DataType <> TYPE_NONE Then
                Set clsVarAsgn = CreateTempTACVar(ctx, DataType)
                ctx.AddCode CreateTAC(TAC_CMD_RETRIEVE, , , clsVarAsgn)
            End If
        
        ' Einfache Zuweisung an Variable
        Case NODE_OP_ASSIGN
            Set clsVarAsgn = ToIStatement(SubNode(0)).GenerateTAC(ctx)
            Set clsVarLeft = ToIStatement(SubNode(1)).GenerateTAC(ctx)
            Set clsVarLeft = DoDeref(ctx, clsVarLeft)
            ctx.AddCode CreateTAC(TAC_CMD_ASSIGN, NodeType, , clsVarAsgn, clsVarLeft)
        
        ' Unre Operatoren haben keinen rechten SubNode
        Case NODE_CONV_BYTE_TO_FLOAT, NODE_CONV_BYTE_TO_INT, _
             NODE_CONV_BYTE_TO_SHORT, NODE_CONV_INT_TO_FLOAT, _
             NODE_CONV_SHORT_TO_FLOAT, NODE_CONV_SHORT_TO_INT, _
             NODE_OP_NEG
            Set clsVarLeft = ToIStatement(SubNode(0)).GenerateTAC(ctx)
            Set clsVarAsgn = CreateTempTACVar(ctx, DataType)
            ctx.AddCode CreateTAC(TAC_CMD_ASSIGN, NodeType, , clsVarAsgn, clsVarLeft)
        
        ' Alle Operatoren mit 2 Faktoren, + - * / ...
        Case Else
            Set clsVarLeft = ToIStatement(SubNode(0)).GenerateTAC(ctx)
            Set clsVarLeft = DoDeref(ctx, clsVarLeft)
            Set clsVarRight = ToIStatement(SubNode(1)).GenerateTAC(ctx)
            Set clsVarRight = DoDeref(ctx, clsVarRight)
            Set clsVarAsgn = CreateTempTACVar(ctx, DataType)
            ctx.AddCode CreateTAC(TAC_CMD_ASSIGN, NodeType, , clsVarAsgn, clsVarLeft, clsVarRight)
            
    End Select
    
    ' Variable, in der das Ergebnis liegt, zurckgeben
    Set IStatement_GenerateTAC = clsVarAsgn
End Function

Private Function DoDeref(ByVal ctx As TACContext, ByVal var As TACVar) As TACVar
    If var.DeRef Then
        Dim clsTemp As TACVar
        Set clsTemp = CreateTempTACVar(ctx, var.DataType)
        ctx.AddCode CreateTAC(TAC_CMD_ASSIGN, NODE_OP_ASSIGN, , clsTemp, var)
        Set DoDeref = clsTemp
    Else
        Set DoDeref = var
    End If
End Function

