VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
  Persistable = 0  'NotPersistable
  DataBindingBehavior = 0  'vbNone
  DataSourceBehavior  = 0  'vbNone
  MTSTransactionMode  = 0  'NotAnMTSObject
END
Attribute VB_Name = "Scanner"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = True
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = False
Option Explicit

Private Const ASC_WHITESPACE        As Byte = 32
Private Const ASC_CARRIAGERETURN    As Byte = 13
Private Const ASC_LINEFEED          As Byte = 10
Private Const ASC_QUOTE             As Byte = 34
Private Const ASC_DOT               As Byte = 46
Private Const ASC_SHARP             As Byte = 35

Public Enum TokenType
    TokenUnknown = 0
    TokenStreamBegin
    TokenStreamEnd
    TokenFunction
    TokenProcedure
    TokenAs
    TokenName
    TokenValue
    TokenString
    TokenBoolTrue
    TokenBoolFalse
    TokenParenthesisLeft    ' (
    TokenParenthesisRight   ' )
    TokenBracketLeft        ' [
    TokenBracketRight       ' ]
    TokenEndOfInstruction   ' vbCrLf
    TokenIf
    TokenThen
    TokenElse
    TokenEnd
    TokenOpEquals           ' ==
    TokenOpAssign           ' =
    TokenOpGreater          ' >
    TokenOpSmaller          ' <
    TokenOpGreaterEqu       ' >=
    TokenOpSmallerEqu       ' =<
    TokenOpUnequal          ' !=
    TokenOpAdd              ' +
    TokenOpSub              ' -
    TokenOpMul              ' *
    TokenOpDiv              ' /
    TokenOpGetPtr           ' @
    TokenOpMod              ' mod
    TokenOpAnd              ' and
    TokenOpOr               ' or
    TokenOpBitAnd           ' bAnd
    TokenOpBitOr            ' bOr
    TokenOpBitXor           ' bXor
    TokenReturn
    TokenFor
    TokenTo
    TokenSeperator          ' ,
    TokenDo
    TokenDim
    TokenReDim
    TokenConst
    TokenPreserve
    TokenRepeat
    TokenWhile
    TokenUntil
    TokenQuote              ' "
    TokenRef
End Enum

Private Type Scanner
    Stream()    As Byte
    Length      As Long
    Position    As Long
    Line        As Long
    LinePos     As Long
    InQuote     As Boolean
End Type

Private Type Token
    ID          As TokenType
    Line        As Long
    Position    As Long
    Content     As String
End Type

Private Scanner As Scanner
Private Token   As Token


Public Property Get TokenID() As TokenType
    TokenID = Token.ID
End Property


Public Property Get TokenContent() As String
    TokenContent = Token.Content
End Property


Public Property Get TokenLine() As Long
    TokenLine = Token.Line
End Property


Public Property Get TokenPosition() As Long
    TokenPosition = Token.Position
End Property


Public Function Match(ByVal typ As TokenType) As Boolean
    If typ = TokenID Then
        Match = True
        Consume
    End If
End Function


Public Function Consume() As String
    Consume = TokenContent
    Token = GetToken()
End Function


Public Sub Reset()
    With Scanner
        .Position = 0
        .LinePos = 0
        .Line = 1
    End With
    
    Token.ID = TokenStreamBegin
End Sub


Public Sub Scan(ByVal Code As String)
    With Scanner
        .Stream = StrConv(Code & vbCrLf, vbFromUnicode)
        .Length = UBound(.Stream) + 1
        .Position = 0
        .LinePos = 0
        .Line = 1
    End With
    
    Token.ID = TokenStreamBegin
End Sub


Private Function GetToken() As Token
    Dim ret As Token
    Dim i   As Long
    
    If Scanner.Position = Scanner.Length Then
        ret.ID = TokenStreamEnd
    Else
        Do While NextChar(i) = ASC_WHITESPACE
            i = i + 1
        Loop
        If NextChar(i) = ASC_SHARP Then
            Do While NextChar(i) <> ASC_CARRIAGERETURN
                i = i + 1
            Loop
        End If
        
        ScannerConsume i
        Scanner.LinePos = Scanner.LinePos + i
        
        Select Case True
            Case NextChar = ASC_QUOTE:
                ret.Content = """"
                ret.ID = TokenQuote
                Scanner.InQuote = Not Scanner.InQuote
                
            Case Scanner.InQuote:
                If NextChar = ASC_CARRIAGERETURN Then
                    Scanner.InQuote = False
                    ret = GetTokenOp()
                Else
                    ret.Content = ParseString()
                    ret.ID = TokenString
                End If
                
            Case IsLetter(NextChar):
                ret = GetTokenName()
                
            Case IsNumber(NextChar):
                ret = GetTokenNumber()
            
            Case Else:
                ret = GetTokenOp()
                
        End Select
    End If
    
    ret.Line = Scanner.Line
    ret.Position = Scanner.LinePos
    Scanner.LinePos = Scanner.LinePos + Len(ret.Content)
    
    If ret.ID = TokenEndOfInstruction Then
        Scanner.Line = Scanner.Line + 1
        Scanner.LinePos = 0
    End If
    
    ScannerConsume Len(ret.Content)
    GetToken = ret
End Function


Private Function GetTokenNumber() As Token
    Dim ret     As Token
    Dim strNum  As String
    Dim blnDot  As Boolean
    Dim i       As Long
    
    Do While IsNumber(NextChar(i)) Or ((NextChar(i) = ASC_DOT) And Not blnDot)
        If NextChar(i) = ASC_DOT Then blnDot = True
        strNum = strNum & ChrW$(NextChar(i))
        i = i + 1
        If Not ScannerValidOffset(i) Then Exit Do
    Loop
    
    ret.Content = strNum
    ret.ID = TokenValue
    
    GetTokenNumber = ret
End Function


Public Function TokenTypeToString(ByVal tk As TokenType) As String
    Select Case tk
        Case TokenEndOfInstruction: TokenTypeToString = "end of instruction"
        Case TokenStreamBegin:      TokenTypeToString = "begin of stream"
        Case TokenStreamEnd:        TokenTypeToString = "end of stream"
        Case TokenUnknown:          TokenTypeToString = "unknown"
        Case TokenAs:               TokenTypeToString = """as"" keyword"
        Case TokenBoolFalse:        TokenTypeToString = """false"" keyword"
        Case TokenBoolTrue:         TokenTypeToString = """true"" keyword"
        Case TokenBracketLeft:      TokenTypeToString = """["" operator"
        Case TokenBracketRight:     TokenTypeToString = """]"" operator"
        Case TokenDim:              TokenTypeToString = """dim"" keyword"
        Case TokenReDim:            TokenTypeToString = """redim"" keyword"
        Case TokenPreserve:         TokenTypeToString = """preserve"" keyword"
        Case TokenDo:               TokenTypeToString = """do"" keyword"
        Case TokenElse:             TokenTypeToString = """else"" keyword"
        Case TokenEnd:              TokenTypeToString = """end"" keyword"
        Case TokenFor:              TokenTypeToString = """for"" keyword"
        Case TokenFunction:         TokenTypeToString = """function"" keyword"
        Case TokenIf:               TokenTypeToString = """if"" keyword"
        Case TokenName:             TokenTypeToString = "identifier"
        Case TokenOpAdd:            TokenTypeToString = """+"" operator"
        Case TokenOpAnd:            TokenTypeToString = """and"" operator"
        Case TokenOpAssign:         TokenTypeToString = "assignment"
        Case TokenOpDiv:            TokenTypeToString = """/"" operator"
        Case TokenOpEquals:         TokenTypeToString = """=="" operator"
        Case TokenOpGetPtr:         TokenTypeToString = """@"" operator"
        Case TokenOpGreater:        TokenTypeToString = """>"" operator"
        Case TokenOpGreaterEqu:     TokenTypeToString = """>="" operator"
        Case TokenOpMod:            TokenTypeToString = """mod"" operator"
        Case TokenOpMul:            TokenTypeToString = """*"" operator"
        Case TokenOpOr:             TokenTypeToString = """or"" operator"
        Case TokenOpSmaller:        TokenTypeToString = """<"" operator"
        Case TokenOpSmallerEqu:     TokenTypeToString = """=<"" operator"
        Case TokenOpSub:            TokenTypeToString = """-"" operator"
        Case TokenOpUnequal:        TokenTypeToString = """!="" operator"
        Case TokenParenthesisLeft:  TokenTypeToString = """("" operator"
        Case TokenParenthesisRight: TokenTypeToString = """)"" operator"
        Case TokenProcedure:        TokenTypeToString = """sub"" keyword"
        Case TokenQuote:            TokenTypeToString = "quote operator"
        Case TokenRef:              TokenTypeToString = """ref"" keyword"
        Case TokenRepeat:           TokenTypeToString = """repeat"" keyword"
        Case TokenReturn:           TokenTypeToString = """return"" keyword"
        Case TokenSeperator:        TokenTypeToString = ""","" operator"
        Case TokenString:           TokenTypeToString = "string"
        Case TokenThen:             TokenTypeToString = """then"" keyword"
        Case TokenTo:               TokenTypeToString = """to"" keyword"
        Case TokenUntil:            TokenTypeToString = """until"" keyword"
        Case TokenValue:            TokenTypeToString = "value"
        Case TokenWhile:            TokenTypeToString = """while"" keyword"
        Case TokenConst:            TokenTypeToString = """const"" keyword"
        Case TokenOpBitAnd:         TokenTypeToString = """bAnd"" keyword"
        Case TokenOpBitOr:          TokenTypeToString = """bOr"" keyword"
        Case TokenOpBitXor:         TokenTypeToString = """bXor"" keyword"
    End Select
End Function


Public Function IDToString() As String
    Select Case TokenID
        Case TokenEndOfInstruction: IDToString = "end of instruction"
        Case TokenStreamBegin:      IDToString = "begin of stream"
        Case TokenStreamEnd:        IDToString = "end of stream"
        Case TokenUnknown:          IDToString = "unknown (" & TokenContent & ")"
        Case Else:                  IDToString = TokenContent
    End Select
End Function


Private Function GetTokenName() As Token
    Dim ret     As Token
    Dim strName As String
    Dim i       As Long
    
    Do While IsLetter(NextChar(i)) Or _
             IsNumber(NextChar(i)) Or _
             IsAllowedChar(NextChar(i))
             
        strName = strName & ChrW$(NextChar(i))
        i = i + 1
        If Not ScannerValidOffset(i) Then Exit Do
    Loop
    
    Select Case LCase$(strName)
        Case "function":    ret.ID = TokenFunction
        Case "sub":         ret.ID = TokenProcedure
        Case "end":         ret.ID = TokenEnd
        Case "for":         ret.ID = TokenFor
        Case "to":          ret.ID = TokenTo
        Case "while":       ret.ID = TokenWhile
        Case "dim":         ret.ID = TokenDim
        Case "redim":       ret.ID = TokenReDim
        Case "const":       ret.ID = TokenConst
        Case "preserve":    ret.ID = TokenPreserve
        Case "ref":         ret.ID = TokenRef
        Case "as":          ret.ID = TokenAs
        Case "if":          ret.ID = TokenIf
        Case "else":        ret.ID = TokenElse
        Case "then":        ret.ID = TokenThen
        Case "mod":         ret.ID = TokenOpMod
        Case "and":         ret.ID = TokenOpAnd
        Case "or":          ret.ID = TokenOpOr
        Case "return":      ret.ID = TokenReturn
        Case "true":        ret.ID = TokenBoolTrue
        Case "false":       ret.ID = TokenBoolFalse
        Case "band":        ret.ID = TokenOpBitAnd
        Case "bor":         ret.ID = TokenOpBitOr
        Case "bxor":        ret.ID = TokenOpBitXor
        Case Else:          ret.ID = TokenName
    End Select
    
    ret.Content = strName
    GetTokenName = ret
End Function


Private Function GetTokenOp() As Token
    Dim ret As Token
    
    Select Case ChrW$(NextChar)
        Case "!"
            Select Case ChrW$(NextChar(1))
                Case "="
                    ret.ID = TokenOpUnequal
                    ret.Content = "!="
                Case Else
                    ret.ID = TokenUnknown
                    ret.Content = "!"
            End Select
        Case "="
            Select Case ChrW$(NextChar(1))
                Case "<"
                    ret.ID = TokenOpSmallerEqu
                    ret.Content = "=<"
                Case "="
                    ret.ID = TokenOpEquals
                    ret.Content = "=="
                Case Else
                    ret.ID = TokenOpAssign
                    ret.Content = "="
            End Select
        Case ">"
            Select Case ChrW$(NextChar(1))
                Case "="
                    ret.ID = TokenOpGreaterEqu
                    ret.Content = ">="
                Case Else
                    ret.ID = TokenOpGreater
                    ret.Content = ">"
            End Select
        Case vbCr:
            If NextChar(1) = ASC_LINEFEED Then
                ret.ID = TokenEndOfInstruction
                ret.Content = vbCrLf
            Else
                ret.ID = TokenUnknown
                ret.Content = vbCr
            End If
        Case "<"
            ret.ID = TokenOpSmaller
            ret.Content = "<"
        Case "@"
            ret.ID = TokenOpGetPtr
            ret.Content = "@"
        Case "+"
            ret.ID = TokenOpAdd
            ret.Content = "+"
        Case "-"
            ret.ID = TokenOpSub
            ret.Content = "-"
        Case "*"
            ret.ID = TokenOpMul
            ret.Content = "*"
        Case "/"
            ret.ID = TokenOpDiv
            ret.Content = "/"
        Case ","
            ret.ID = TokenSeperator
            ret.Content = ","
        Case "("
            ret.ID = TokenParenthesisLeft
            ret.Content = "("
        Case ")"
            ret.ID = TokenParenthesisRight
            ret.Content = ")"
        Case "["
            ret.ID = TokenBracketLeft
            ret.Content = "["
        Case "]"
            ret.ID = TokenBracketRight
            ret.Content = "]"
        Case Else
            ret.ID = TokenUnknown
            ret.Content = ChrW$(NextChar)
    End Select
    
    GetTokenOp = ret
End Function


Private Function ParseString() As String
    Dim i   As Long
    Dim ret As String
    
    i = 0
    
    Do While (NextChar(i) <> ASC_QUOTE) And _
             (NextChar(i) <> ASC_CARRIAGERETURN) And _
             ScannerValidOffset(i)
        
        ret = ret & ChrW$(NextChar(i))
        i = i + 1
    Loop
    
    ParseString = ret
End Function


Private Function ScannerValidOffset(ByVal Bytes As Long) As Boolean
    ScannerValidOffset = (Scanner.Position + Bytes < Scanner.Length)
End Function


Private Sub ScannerConsume(ByVal Bytes As Long)
    Scanner.Position = Scanner.Position + Bytes
End Sub


Private Property Get NextChar(Optional ByVal Offset As Long = 0) As Byte
    NextChar = Scanner.Stream(Scanner.Position + Offset)
End Property


Private Function IsAllowedChar(ByVal bt As Byte) As Boolean
    IsAllowedChar = (bt = 95)
End Function


Private Function IsLetter(ByVal bt As Byte) As Boolean
    IsLetter = ((bt >= 65) And (bt <= 90)) Or ((bt >= 97) And (bt <= 122))
End Function


Private Function IsNumber(ByVal bt As Byte) As Boolean
    IsNumber = (bt >= 48) And (bt <= 57)
End Function
