VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
  Persistable = 0  'NotPersistable
  DataBindingBehavior = 0  'vbNone
  DataSourceBehavior  = 0  'vbNone
  MTSTransactionMode  = 0  'NotAnMTSObject
END
Attribute VB_Name = "StreamMP3"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = True
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = True
Option Explicit


''' MPEG Audio Decoder (mad [GPL]) Wrapper
''
'' - MPEG 1/2/2.5 Layer I/II/III (+ FreeFormat)
'' - VBR (XING, VBRI)
'' - ID3v1/2 Tags ber Tagging Modul von Mathias Kunter (s. ID3Tags.bas)
'
' Arne Elster 2007/08/09
'
'
' TODO: MP3s mit mehreren ID3v2 Tags untersttzen (?!?!?)


Implements IPlugin
Implements IPluginStream


Private Const MyDescription As String = "libmad MP3 Decoder"
Private Const MyExtensions As String = "mp3;mp2"


Private Declare Function CreateFile Lib "Kernel32.dll" _
Alias "CreateFileA" ( _
    ByVal lpFileName As String, _
    ByVal dwDesiredAccess As Long, ByVal dwShareMode As Long, _
    lpSecurityAttributes As Any, ByVal dwCreationDisposition As Long, _
    ByVal dwFlagsAndAttributes As Long, ByVal hTemplateFile As Long _
) As Long

Private Declare Function ReadFile Lib "kernel32" ( _
    ByVal hFile As Long, lpBuffer As Any, _
    ByVal nNumberOfBytesToRead As Long, lpNumberOfBytesRead As Long, _
    ByVal lpOverlapped As Any _
) As Long

Private Declare Function WriteFile Lib "kernel32" ( _
    ByVal hFile As Long, lpBuffer As Any, _
    ByVal nNumberOfBytesToWrite As Long, lpNumberOfBytesWritten As Long, _
    ByVal lpOverlapped As Any _
) As Long

Private Declare Function SetFilePointer Lib "kernel32" ( _
    ByVal hFile As Long, _
    ByVal lDistanceToMove As Long, ByVal lpDistanceToMoveHigh As Long, _
    ByVal dwMoveMethod As Long _
) As Long

Private Declare Function GetFileSize Lib "kernel32" ( _
    ByVal hFile As Long, ByVal lpFileSizeHigh As Long _
) As Long

Private Declare Function CloseHandle Lib "kernel32" ( _
    ByVal hObject As Long _
) As Long

Private Enum FILE_OPEN_METHOD
    CREATE_NEW = 1
    CREATE_ALWAYS = 2
    OPEN_EXISTING = 3
    OPEN_ALWAYS = 4
End Enum

Private Enum FILE_SHARE_RIGHTS
    FILE_SHARE_READ = &H1
    FILE_SHARE_WRITE = &H2
End Enum

Private Enum FILE_ACCESS_RIGHTS
    GENERIC_READ = &H80000000
    GENERIC_WRITE = &H40000000
End Enum

Private Enum SEEK_METHOD
    FILE_BEGIN = 0
    FILE_CURRENT = 1
    FILE_END = 2
End Enum

Private Const INVALID_HANDLE As Long = -1

Private Type hFile
    handle      As Long
    path        As String
End Type

' ***************************************************

Private Declare Sub CpyMem Lib "kernel32" Alias "RtlMoveMemory" ( _
    pDst As Any, pSrc As Any, ByVal cb As Long _
)

' ***************************************************
' ** MAD Timer
' *********************

Private Declare Function mad_timer_compare Lib "libmad.dll" ( _
    timer1 As mad_timer_t, timer2 As mad_timer_t _
) As Long

Private Declare Sub mad_timer_negate Lib "libmad.dll" ( _
    timer As mad_timer_t _
)

Private Declare Function mad_timer_abs Lib "libmad.dll" ( _
    timr As mad_timer_t _
) As mad_timer_t

Private Declare Sub mad_timer_set Lib "libmad.dll" ( _
    timer As mad_timer_t, ByVal seconds As Long, _
    ByVal numer As Long, ByVal denom As Long _
)

Private Declare Sub mad_timer_add_api Lib "libmad.dll" _
Alias "mad_timer_add" ( _
    timer As mad_timer_t, ByVal seconds As Long, ByVal fraction As Long _
)

Private Declare Sub mad_timer_multiply Lib "libmad.dll" ( _
    timer As mad_timer_t, ByVal scalar As Long _
)

Private Declare Function mad_timer_count_api Lib "libmad.dll" _
Alias "mad_timer_count" ( _
    ByVal seconds As Long, ByVal fraction As Long, ByVal units As mad_units _
) As Long

Private Declare Function mad_timer_fraction Lib "libmad.dll" ( _
    timer As mad_timer_t, ByVal denom As Long _
) As Long

Private Declare Sub mad_timer_string Lib "libmad.dll" ( _
    timer As mad_timer_t, ByVal dest As String, _
    ByVal format As String, ByVal units As mad_units, _
    ByVal fracunits As mad_units, ByVal subparts As Long _
)

Private Type mad_timer_t
    seconds     As Long
    fraction    As Long
End Type

Private Const MAD_TIMER_RESOLUTION As Long = 352800000

Private Enum mad_units
    MAD_UNITS_HOURS = -2
    MAD_UNITS_MINUTES = -1
    MAD_UNITS_SECONDS = 0
    
    MAD_UNITS_DECISECONDS = 10
    MAD_UNITS_CENTISECONDS = 100
    MAD_UNITS_MILLISECONDS = 1000
    
    MAD_UNITS_8000_HZ = 8000
    MAD_UNITS_11025_HZ = 11025
    MAD_UNITS_12000_HZ = 12000
    MAD_UNITS_16000_HZ = 16000
    MAD_UNITS_22050_HZ = 22050
    MAD_UNITS_24000_HZ = 24000
    MAD_UNITS_32000_HZ = 32000
    MAD_UNITS_44100_HZ = 44100
    MAD_UNITS_48000_HZ = 48000
    
    MAD_UNITS_24_FPS = 24
    MAD_UNITS_25_FPS = 25
    MAD_UNITS_30_FPS = 30
    MAD_UNITS_48_FPS = 48
    MAD_UNITS_50_FPS = 50
    MAD_UNITS_60_FPS = 60
    
    MAD_UNITS_75_FPS = 75
    
    MAD_UNITS_23_976_FPS = -24
    MAD_UNITS_24_975_FPS = -25
    MAD_UNITS_29_97_FPS = -30
    MAD_UNITS_47_952_FPS = -48
    MAD_UNITS_49_95_FPS = -50
    MAD_UNITS_59_94_FPS = -60
End Enum

' ***************************************************
' ** MAD Bit Ptr
' *********************

Private Declare Sub mad_bit_init Lib "libmad.dll" ( _
    bitptr As mad_bitptr, pByte As Byte _
)

Private Declare Function mad_bit_length Lib "libmad.dll" ( _
    bitbegin As mad_bitptr, bitend As mad_bitptr _
) As Long

Private Declare Function mad_bit_nextbyte Lib "libmad.dll" ( _
    bitptr As mad_bitptr _
) As Long

Private Declare Sub mad_bit_skip Lib "libmad.dll" ( _
    bitptr As mad_bitptr, ByVal Length As Long _
)

Private Declare Function mad_bit_read Lib "libmad.dll" ( _
    bitptr As mad_bitptr, ByVal Length As Long _
) As Long

Private Declare Function mad_bit_crc Lib "libmad.dll" ( _
    bitptr As mad_bitptr, ByVal Length As Long, ByVal Init As Integer _
) As Integer

Private Type mad_bitptr
    pByte       As Long
    Cache       As Integer
    BytesLeft   As Integer
End Type

' ***************************************************
' ** MAD Decoder
' *********************

Private Declare Sub mad_decoder_init Lib "libmad.dll" ( _
    decoder As mad_decoder, pData As Any, _
    ByVal pInput_func As Long, ByVal pHeader_func As Long, _
    ByVal pFilter_func As Long, ByVal poutput_func As Long, _
    ByVal pError_func As Long, ByVal pMessage_func As Long _
)

Private Declare Function mad_decoder_finish Lib "libmad.dll" ( _
    decoder As mad_decoder _
) As Long

Private Declare Function mad_decoder_run Lib "libmad.dll" ( _
    decoder As mad_decoder, ByVal mode As mad_decoder_mode _
) As Long

Private Declare Function mad_decoder_message Lib "libmad.dll" ( _
    decoder As mad_decoder, message As Any, Length As Long _
) As Long

Private Declare Sub mad_sample_scale Lib "libmad.dll" ( _
    samples_left As Any, samples_right As Any, output As Any, _
    ByVal Length As Long, ByVal Channels As Long _
)

Private Enum mad_decoder_mode
    MAD_DECODER_MODE_SYNC = 0
    MAD_DECODER_MODE_ASYNC
End Enum

Private Enum mad_flow
    MAD_FLOW_CONTINUE = &H0
    MAD_FLOW_STOP = &H10
    MAD_FLOW_BREAK = &H11
    MAD_FLOW_IGNORE = &H20
End Enum

Private Type mad_async
    pid             As Long
    in              As Long
    out             As Long
End Type

Private Type mad_decoder
    mode            As mad_decoder_mode
    options         As Long
    async           As mad_async
    pSync           As Long
    cbData          As Long
    pInput_func     As Long
    pHeader_func    As Long
    pFilter_func    As Long
    poutput_func    As Long
    pError_func     As Long
    pMessage_func   As Long
End Type

' ***************************************************
' ** MAD Frame
' *********************

Private Declare Sub mad_header_init Lib "libmad.dll" (hdr As mad_header)

Private Declare Sub mad_frame_init Lib "libmad.dll" (frm As mad_frame)

Private Declare Sub mad_frame_finish Lib "libmad.dll" (frm As mad_frame)

Private Declare Function mad_header_decode Lib "libmad.dll" ( _
    hdr As mad_header, strm As mad_stream _
) As Long

Private Declare Function mad_frame_decode Lib "libmad.dll" ( _
    frm As mad_frame, strm As mad_stream _
) As Long

Private Declare Sub mad_frame_mute Lib "libmad.dll" (frm As mad_frame)

Private Enum mad_layer
    MAD_LAYER_I = 1
    MAD_LAYER_II = 2
    MAD_LAYER_III = 3
End Enum

Private Enum mad_mode
    MAD_MODE_SINGLE_CHANNEL = 0
    MAD_MODE_DUAL_CHANNEL = 1
    MAD_MODE_JOINT_STEREO = 2
    MAD_MODE_STEREO = 3
End Enum

Private Enum mad_emphasis
    MAD_EMPHASIS_NONE = 0
    MAD_EMPHASIS_50_15_US = 1
    MAD_EMPHASIS_CCITT_J_17 = 3
    MAD_EMPHASIS_RESERVED = 2
End Enum

Private Type mad_header
    Layer           As mad_layer
    mode            As mad_mode
    mode_extension  As Long
    Emphasis        As mad_emphasis
    Bitrate         As Long
    samplerate      As Long
    crc_check       As Integer
    crc_target      As Integer
    Flags           As Long
    private_bits    As Long
    Duration        As mad_timer_t
End Type

Private Type mad_frame
    header          As mad_header
    options         As Long
    sbsample(2303)  As Long
    pOverlap        As Long
End Type

Private Const MAD_FLAG_NPRIVATE_III As Long = &H7
Private Const MAD_FLAG_INCOMPLETE   As Long = &H8

Private Const MAD_FLAG_PROTECTION   As Long = &H10
Private Const MAD_FLAG_COPYRIGHT    As Long = &H20
Private Const MAD_FLAG_ORIGINAL     As Long = &H40
Private Const MAD_FLAG_PADDING      As Long = &H80

Private Const MAD_FLAG_I_STEREO     As Long = &H100
Private Const MAD_FLAG_MS_STEREO    As Long = &H200
Private Const MAD_FLAG_FREEFORMAT   As Long = &H400

Private Const MAD_FLAG_LSF_EXT      As Long = &H1000
Private Const MAD_FLAG_MC_EXT       As Long = &H2000
Private Const MAD_FLAG_MPEG_2_5_EXT As Long = &H4000

Private Enum mad_private
    MAD_PRIVATE_HEADER = &H100
    MAD_PRIVATE_III = &H1F
End Enum


' ***************************************************
' ** MAD Stream
' *********************

Private Declare Sub mad_stream_init Lib "libmad.dll" (Stream As mad_stream)

Private Declare Sub mad_stream_finish Lib "libmad.dll" (Stream As mad_stream)

Private Declare Sub mad_stream_buffer Lib "libmad.dll" ( _
    Stream As mad_stream, buffer As Any, ByVal Length As Long _
)

Private Declare Sub mad_stream_skip Lib "libmad.dll" ( _
    Stream As mad_stream, ByVal Length As Long _
)

Private Declare Function mad_stream_sync Lib "libmad.dll" ( _
    Stream As mad_stream _
) As Long

Private Declare Function mad_stream_errorstr Lib "libmad.dll" ( _
    Stream As mad_stream _
) As Long

Private Const MAD_BUFFER_GUARD As Long = 8
Private Const MAD_BUFFER_MDLEN As Long = (511 + 2048 + MAD_BUFFER_GUARD)

Private Enum mad_error
    MAD_ERROR_NONE = &H0
    
    MAD_ERROR_BUFLEN = &H1
    MAD_ERROR_BUFPTR = &H2
    MAD_ERROR_NOMEM = &H31
    
    MAD_ERROR_LOSTSYNC = &H101
    MAD_ERROR_BADLAYER = &H102
    MAD_ERROR_BADBITRATE = &H103
    MAD_ERROR_BADSAMPLERATE = &H104
    MAD_ERROR_BADEMPHASIS = &H105
    
    MAD_ERROR_BADCRC = &H201
    MAD_ERROR_BADBITALLOC = &H211
    MAD_ERROR_BADSCALEFACTOR = &H221
    MAD_ERROR_BADMODE = &H222
    MAD_ERROR_BADFRAMELEN = &H231
    MAD_ERROR_BADBIGVALUES = &H232
    MAD_ERROR_BADBLOCKTYPE = &H233
    MAD_ERROR_BADSCFSI = &H234
    MAD_ERROR_BADDATAPTR = &H235
    MAD_ERROR_BADPART3LEN = &H236
    MAD_ERROR_BADHUFFTABLE = &H237
    MAD_ERROR_BADHUFFDATA = &H238
    MAD_ERROR_BADSTEREO = &H239
End Enum

Private Type mad_stream
    buffer      As Long
    bufend      As Long
    skiplen     As Long
    sync        As Long
    freerate    As Long
    this_frame  As Long
    next_frame  As Long
    ptr         As mad_bitptr
    anc_ptr     As mad_bitptr
    anc_bitlen  As Long
    main_data   As Long
    md_len      As Long
    options     As Long
    nError      As mad_error
End Type

Private Enum mad_option
    MAD_OPTION_IGNORECRC = &H1
    MAD_OPTION_HALFSAMPLERATE = &H2
End Enum

' Kein Frame sollte grer als BufferSize sein
Private Const MADBufferSize         As Long = 4096


' ***************************************************
' ** MAD Synth
' *********************

Private Declare Sub mad_synth_init Lib "libmad.dll" (synth As mad_synth)

Private Declare Sub mad_synth_mute Lib "libmad.dll" (synth As mad_synth)

Private Declare Sub mad_synth_frame Lib "libmad.dll" ( _
    synth As mad_synth, Frame As mad_frame _
)

Private Type mad_pcm
    samplerate      As Long
    Channels        As Integer
    Length          As Integer
    samples(2303)   As Long
End Type

Private Type mad_synth
    filter(1023)    As Long
    phase           As Long
    pcm             As mad_pcm
End Type

Private Const MAD_PCM_CHANNEL_SINGLE        As Long = 0
Private Const MAD_PCM_CHANNEL_DUAL_1        As Long = 0
Private Const MAD_PCM_CHANNEL_DUAL_2        As Long = 1
Private Const MAD_PCM_CHANNEL_STEREO_LEFT   As Long = 0
Private Const MAD_PCM_CHANNEL_STEREO_RIGHT  As Long = 1

''''''''''''''''''''''''''''''''''''''''
''''''''''''''''''''''''''''''''''''''''

Private Enum ID3v2Flags
    ID3_Unsync = (2 ^ 7)
    ID3_ExtHdr = (2 ^ 6)
    ID3_Experm = (2 ^ 5)
    ID3_Footer = (2 ^ 4)
End Enum

Private Enum VBRMode
    VBRNone
    VBRXing
    VBRVBRI
End Enum

Private Enum XingFlags
    XING_Frames = &H1
    XING_Bytes = &H2
    XING_TOC = &H4
    XING_Scale = &H8
End Enum

Private Type XingHeader
    Flags                   As XingFlags
    bytes                   As Long
    Frames                  As Long
    vbrscale                As Long
    Toc(99)                 As Byte
End Type

Private Type VBRIHeader
    VbriVersion             As Long
    VbriDelay               As Single
    VbriQuality             As Long
    VbrIPluginStreamBytes         As Long
    VbrIPluginStreamFrames        As Long
    VbriTableSize           As Long
    VbriTableScale          As Long
    VbriEntryBytes          As Long
    VbriEntryFrames         As Long
    VbriTable()             As Long
End Type

Private Type MPEGInfo
    samplerate              As Long
    Channels                As Integer
    CBRBitrate              As Long
    VBRBitrate              As Long
    Layer                   As Integer
    MPEGVersion             As Single
    Copyrighted             As Boolean
    Original                As Boolean
    Protected               As Boolean
    FreeFormat              As Boolean
    Padding                 As Boolean
    Emphasis                As Boolean
    SamplesPerFrame         As Long
    FrameSize               As Long
    Frames                  As Long
    bytes                   As Long
    ID3v1                   As Boolean
    ID3v2                   As Boolean
    Tags                    As ID3Tag
    VBR                     As VBRMode
    VBRXing                 As XingHeader
    VBRVBRI                 As VBRIHeader
    InfoHeaderValid         As Boolean
    InfoHeader              As XingHeader
End Type

Private m_hFile             As hFile
Private m_lngFirstFrame     As Long
Private m_udtInfo           As MPEGInfo

Private m_mad_stream        As mad_stream
Private m_mad_frame         As mad_frame
Private m_mad_synth         As mad_synth
Private m_mad_header        As mad_header
Private m_mad_timer         As mad_timer_t

Private m_clsInfo           As StreamTags

Private m_blnEOF            As Boolean

Private m_sngSamples()      As Single
Private m_lngSampleCount    As Long
Private m_lngSamplesUsed    As Long

Private Function IPluginStream_CreateNewStream() As IPluginStream
    Set IPluginStream_CreateNewStream = New StreamMP3
End Function

Private Property Get IPluginStream_Description() As String
    IPluginStream_Description = MyDescription
End Property

Private Property Get IPluginStream_Extensions() As String()
    IPluginStream_Extensions = Split(MyExtensions, ";")
End Property

Private Function IPluginStream_StreamClose() As STREAM_RESULT
    Dim emptyinfo As MPEGInfo
    
    If m_hFile.handle = INVALID_HANDLE Then
        IPluginStream_StreamClose = RESULT_UNKNOWN_ERROR
        Exit Function
    End If

    FileClose m_hFile
    
    mad_stream_finish m_mad_stream
    mad_synth_finish m_mad_synth
    mad_frame_finish m_mad_frame
    mad_header_finish m_mad_header
    mad_timer_reset m_mad_timer
    
    ' Tagnderungen erst nach Schlieen wirksam machen,
    ' um flssige Wiedergabe zu gewhrleisten
    Dim i As Long
    For i = 1 To m_clsInfo.TagCount
        If m_clsInfo.Tag(i).Changed Or m_clsInfo.Changed Then
            Dim udtID3 As ID3Tag
            With udtID3
                If m_clsInfo.TagExists("artist") Then _
                    .Artist = m_clsInfo.Tag("artist").value
                If m_clsInfo.TagExists("album") Then _
                    .Album = m_clsInfo.Tag("album").value
                If m_clsInfo.TagExists("title") Then _
                    .Title = m_clsInfo.Tag("title").value
                If m_clsInfo.TagExists("comment") Then _
                    .Comment = m_clsInfo.Tag("comment").value
                If m_clsInfo.TagExists("genre") Then _
                    .Genre = m_clsInfo.Tag("genre").value
                If m_clsInfo.TagExists("tracknr") Then _
                    .TrackNr = m_clsInfo.Tag("tracknr").value
                If m_clsInfo.TagExists("composer") Then _
                    .Composer = m_clsInfo.Tag("composer").value
                If m_clsInfo.TagExists("year") Then _
                    .SongYear = m_clsInfo.Tag("year").value
            End With
            WriteID3v1 m_hFile.path, udtID3, True
            WriteID3v2 m_hFile.path, udtID3, VERSION_2_3, False
            Exit For
        End If
    Next
    
    Set m_clsInfo = New StreamTags
    m_udtInfo = emptyinfo
    m_blnEOF = False
    
    IPluginStream_StreamClose = RESULT_SUCCESS
End Function

Private Property Get IPluginStream_StreamInfo() As StreamTags
    Set IPluginStream_StreamInfo = m_clsInfo
End Property

Private Function IPluginStream_StreamOpen(ByVal strPath As String) As STREAM_RESULT
    IPluginStream_StreamClose
    
    m_hFile = FileOpen(strPath, GENERIC_READ, _
                       FILE_SHARE_READ Or FILE_SHARE_WRITE, _
                       OPEN_EXISTING)

    If m_hFile.handle = INVALID_HANDLE Then
        IPluginStream_StreamOpen = RESULT_CANT_OPEN
        Exit Function
    End If

    mad_header_init m_mad_header
    mad_frame_init m_mad_frame
    mad_synth_init m_mad_synth
    mad_stream_init m_mad_stream
    mad_timer_reset m_mad_timer
    
    If Not FindFirstFrame(m_lngFirstFrame, m_udtInfo) Then
        IPluginStream_StreamClose
        IPluginStream_StreamOpen = RESULT_INVALID_FORMAT
    Else
        If m_udtInfo.ID3v2 Then
            ReadID3v2 strPath, m_udtInfo.Tags
        ElseIf m_udtInfo.ID3v1 Then
            ReadID3v1 strPath, m_udtInfo.Tags
        End If
        
        FileSeek m_hFile, m_lngFirstFrame, FILE_BEGIN
        FillInfo
        FillBuffer
        
        IPluginStream_StreamOpen = RESULT_SUCCESS
    End If
End Function

Private Sub FillInfo()
    With m_udtInfo
        If Len(.Tags.Artist) > 0 Then m_clsInfo.AddTag "artist", .Tags.Artist
        If Len(.Tags.Album) > 0 Then m_clsInfo.AddTag "album", .Tags.Album
        If Len(.Tags.Title) > 0 Then m_clsInfo.AddTag "title", .Tags.Title
        If Len(.Tags.Genre) > 0 Then m_clsInfo.AddTag "genre", .Tags.Genre
        If Len(.Tags.Comment) > 0 Then m_clsInfo.AddTag "comment", .Tags.Comment
        If Len(.Tags.TrackNr) > 0 Then m_clsInfo.AddTag "tracknr", .Tags.TrackNr
        If Len(.Tags.Composer) > 0 Then m_clsInfo.AddTag "composer", .Tags.Composer
        If Len(.Tags.SongYear) > 0 Then m_clsInfo.AddTag "year", .Tags.SongYear
        
        m_clsInfo.AddInfo "id3v1", IIf(.ID3v1, "ja", "nein")
        m_clsInfo.AddInfo "id3v2", IIf(.ID3v2, "ja", "nein")
        m_clsInfo.AddInfo "channels", .Channels
        m_clsInfo.AddInfo "samplerate", .samplerate & " Hz"
        m_clsInfo.AddInfo "frames", .Frames
        m_clsInfo.AddInfo "layer", Choose(.Layer, "I", "II", "III")
        m_clsInfo.AddInfo "mpegversion", .MPEGVersion
        m_clsInfo.AddInfo "VBR", IIf(.VBR, "ja", "nein")
        m_clsInfo.AddInfo "bitrate", (.VBRBitrate \ 1000) & " kbit/s"
        m_clsInfo.AddInfo "copyright", IIf(.Copyrighted, "ja", "nein")
        m_clsInfo.AddInfo "original", IIf(.Original, "ja", "nein")
        m_clsInfo.AddInfo "protected", IIf(.Protected, "ja", "nein")
    End With
End Sub

Private Function GetFrame(Frame() As Integer, lngReturned As Long) As STREAM_RESULT
    Const BUFFERSIZE As Long = 4& * 1024&

    Dim btBuffer(BUFFERSIZE - 1)    As Byte
    Dim lngRead                     As Long
    Dim lngConsumed                 As Long
    Dim lngBufSize                  As Long
    Dim blnValidFrame               As Boolean
    
    lngReturned = 0
    
    If m_blnEOF Then
        GetFrame = RESULT_END_OF_STREAM
        Exit Function
    End If
    
    lngRead = FileRead(m_hFile, VarPtr(btBuffer(0)), BUFFERSIZE)
    lngBufSize = lngRead
    
    If lngRead = 0 Then
        m_blnEOF = True
        GetFrame = RESULT_END_OF_STREAM
        Exit Function
    End If
    
    mad_stream_buffer m_mad_stream, btBuffer(0), lngBufSize
    
    blnValidFrame = True
    
    Do While mad_frame_decode(m_mad_frame, m_mad_stream) <> 0
        If Not ReSync() Then
            blnValidFrame = False
            Exit Do
        End If

        lngRead = FileRead(m_hFile, VarPtr(btBuffer(0)), lngBufSize)
        If lngRead = 0 Then
            blnValidFrame = False
            Exit Do
        End If
        
        mad_stream_buffer m_mad_stream, btBuffer(0), lngRead
    Loop
    
    If Not blnValidFrame Then
        GetFrame = RESULT_END_OF_STREAM
        m_blnEOF = True
        Exit Function
    End If

    lngConsumed = m_mad_stream.next_frame - m_mad_stream.this_frame
    FileSeek m_hFile, FilePosition(m_hFile) - lngBufSize + lngConsumed, FILE_BEGIN

    mad_timer_add m_mad_timer, m_mad_frame.header.Duration
    mad_synth_frame m_mad_synth, m_mad_frame

    With m_mad_synth.pcm
        ReDim Frame(.Length * .Channels - 1) As Integer
        mad_sample_scale .samples(0), .samples(1152), Frame(0), .Length, .Channels
    End With
    
    lngReturned = UBound(Frame) + 1
    GetFrame = RESULT_SUCCESS
End Function

Private Function ExtractVBRIHeader( _
    ByRef Frame() As Byte, _
    ByVal Position As Long, _
    ByRef header As VBRIHeader _
) As Boolean

    Dim TableLength As Long
    Dim i           As Long

    Position = Position + 4
    
    header.VbriVersion = VBRIExtract(Frame, Position, 2)
    header.VbriDelay = VBRIExtract(Frame, Position, 2)
    header.VbriQuality = VBRIExtract(Frame, Position, 2)
    header.VbrIPluginStreamBytes = VBRIExtract(Frame, Position, 4)
    header.VbrIPluginStreamFrames = VBRIExtract(Frame, Position, 4)
    header.VbriTableSize = VBRIExtract(Frame, Position, 2)
    header.VbriTableScale = VBRIExtract(Frame, Position, 2)
    header.VbriEntryBytes = VBRIExtract(Frame, Position, 2)
    header.VbriEntryFrames = VBRIExtract(Frame, Position, 2)
    
    ReDim header.VbriTable(header.VbriTableSize + 1) As Long
    
    With header
        For i = 0 To header.VbriTableSize
            .VbriTable(i) = VBRIExtract(Frame, Position, .VbriEntryBytes) _
                            * .VbriTableScale
        Next
    End With
    
    ExtractVBRIHeader = True
End Function

Private Function ExtractXingHeader( _
    ByRef Frame() As Byte, _
    ByVal Position As Long, _
    ByRef header As XingHeader _
) As Boolean

    Dim foff  As Long

    header.Flags = XingExtractI4(Frame, Position + 4)
    
    If (header.Flags And XING_Frames) Then
        header.Frames = XingExtractI4(Frame, Position + 8 + foff)
        foff = foff + 4
    End If
    
    If (header.Flags And XING_Bytes) Then
        header.bytes = XingExtractI4(Frame, Position + 8 + foff)
        foff = foff + 4
    End If
    
    If (header.Flags And XING_TOC) Then
        CpyMem header.Toc(0), Frame(Position + 8 + foff), 100
        foff = foff + 100
    End If
    
    If (header.Flags And XING_Scale) Then
        header.vbrscale = XingExtractI4(Frame, Position + 8 + foff)
        foff = foff + 4
    End If
    
    ExtractXingHeader = True
End Function

Private Function VBRIExtract( _
    buf() As Byte, _
    ByRef Position As Long, _
    ByVal bytes As Long _
) As Long

    Dim i As Long, b As Long, number As Long
    
    For i = 0 To bytes - 1
        b = bytes - 1 - i
        number = number Or ((buf(Position + i) And &HFF) * (2 ^ (8 * b)))
    Next
    
    Position = Position + bytes
    VBRIExtract = number
End Function

Private Function XingExtractI2( _
    buf() As Byte, _
    ByVal Position As Long _
) As Integer

    Dim X       As Integer
    Dim Bt(1)   As Byte
    
    Bt(0) = buf(Position + 1)
    Bt(1) = buf(Position + 0)
    
    CpyMem X, Bt(0), 2
    XingExtractI2 = X
End Function

Private Function XingExtractI4( _
    buf() As Byte, _
    ByVal Position As Long _
) As Long

    Dim X       As Long
    Dim Bt(3)   As Byte

    Bt(0) = buf(Position + 3)
    Bt(1) = buf(Position + 2)
    Bt(2) = buf(Position + 1)
    Bt(3) = buf(Position + 0)
    
    CpyMem X, Bt(0), 4
    XingExtractI4 = X
End Function

Private Function FindFirstFrame( _
    ByRef Position As Long, _
    ByRef info As MPEGInfo _
) As Boolean
    
    Dim btBuffer(MADBufferSize - 1) As Byte
    Dim lngRead                     As Long
    Dim lngBufSize                  As Long
    Dim lngConsumed                 As Long
    Dim lngFrameSize                As Long
    Dim lngReadAbs                  As Long
    Dim i                           As Long
    Dim blnValid                    As Boolean
    Dim lngID3v2Size                As Long
    Dim lngID3v2Offs                As Long

    info.ID3v1 = FindID3v1(0, 0)

    If FindID3v2(lngID3v2Offs, lngID3v2Size) Then
        info.ID3v2 = True
        FileSeek m_hFile, lngID3v2Offs + lngID3v2Size, FILE_BEGIN
        lngReadAbs = FilePosition(m_hFile)
    Else
        FileSeek m_hFile, 0, FILE_BEGIN
    End If
    
    lngRead = FileRead(m_hFile, VarPtr(btBuffer(0)), MADBufferSize)
    lngBufSize = lngRead
    mad_stream_buffer m_mad_stream, btBuffer(0), lngBufSize
    
    Do
        If mad_header_decode(m_mad_header, m_mad_stream) <> 0 Then
            If m_mad_stream.nError <> MAD_ERROR_NONE Then
                If Not MAD_RECOVERABLE(m_mad_stream.nError) Then
                    If m_mad_stream.nError = MAD_ERROR_BUFLEN Then
                        lngConsumed = m_mad_stream.bufend _
                                      - m_mad_stream.next_frame _
                                      - MAD_BUFFER_GUARD
                        
                        If lngConsumed > 0 Then
                            CpyMem btBuffer(0), _
                                   ByVal m_mad_stream.next_frame, _
                                   lngConsumed
                                   
                            lngReadAbs = lngReadAbs - MAD_BUFFER_GUARD
                        Else
                            FileSeek m_hFile, FilePosition(m_hFile) _
                                              - lngBufSize _
                                              + (lngBufSize / 2), _
                                     FILE_BEGIN
                                     
                            lngReadAbs = lngReadAbs - (lngBufSize / 2)
                        End If
                        
                        lngBufSize = lngConsumed
                        
                        lngRead = FileRead(m_hFile, _
                                           VarPtr(btBuffer(lngBufSize)), _
                                           (m_mad_stream.bufend _
                                            - m_mad_stream.buffer) _
                                           - lngConsumed)

                        lngBufSize = lngBufSize + lngRead
                        lngReadAbs = lngReadAbs + lngRead
                        
                        If lngRead = 0 Then Exit Do
                        
                        mad_stream_buffer m_mad_stream, _
                                          btBuffer(0), _
                                          lngBufSize
                    Else
                        Exit Do
                    End If
                End If
            End If
        Else
            blnValid = True
            Exit Do
        End If
    Loop
    
    If Not blnValid Then Exit Function
    
    Position = lngReadAbs + (m_mad_stream.this_frame - m_mad_stream.buffer)
    
    FileSeek m_hFile, Position, FILE_BEGIN
    FileRead m_hFile, VarPtr(btBuffer(0)), MADBufferSize

    info.VBR = VBRNone
    
    For i = 0 To MADBufferSize - 5
        Select Case BytesToAnsi(btBuffer, i, 4)
        
            Case "Info"     ' CBR Info Header
                If ExtractXingHeader(btBuffer, i, info.InfoHeader) Then
                    info.InfoHeaderValid = True
                End If
                
                Exit For
        
            Case "Xing"     ' Xing VBR Header
                If ExtractXingHeader(btBuffer, i, info.VBRXing) Then
                    info.VBR = VBRXing
                End If

                Exit For
                
            Case "VBRI"     ' Frauenhofer VBR Header
                If ExtractVBRIHeader(btBuffer, i, info.VBRVBRI) Then
                    info.VBR = VBRVBRI
                End If
                
                Exit For

        End Select
    Next

'    ' VBR Header, ersten Frame berspringen
'    If info.VBR <> VBRNone Or info.InfoHeaderValid Then
'        m_lngFirstFrame = m_lngFirstFrame + (m_mad_stream.next_frame - m_mad_stream.this_frame)
'    End If

    ' MPEG Bytes
    If info.InfoHeaderValid And (info.InfoHeader.Flags And XING_Bytes) Then
        info.bytes = info.InfoHeader.bytes
    ElseIf info.VBR = VBRXing And (info.VBRXing.Flags And XING_Bytes) Then
        info.bytes = info.VBRXing.bytes
    ElseIf info.VBR = VBRVBRI Then
        info.bytes = info.VBRVBRI.VbrIPluginStreamBytes
    Else
        If info.ID3v1 Then
            info.bytes = FileLength(m_hFile) - m_lngFirstFrame - 128
        Else
            info.bytes = FileLength(m_hFile) - m_lngFirstFrame
        End If
    End If
        
    info.samplerate = m_mad_header.samplerate
    info.Channels = MAD_NCHANNELS(m_mad_header)
    info.CBRBitrate = m_mad_header.Bitrate
    info.Layer = m_mad_header.Layer

    info.Emphasis = m_mad_header.Emphasis <> MAD_EMPHASIS_NONE
    info.Copyrighted = m_mad_header.Flags And MAD_FLAG_COPYRIGHT
    info.Original = m_mad_header.Flags And MAD_FLAG_ORIGINAL
    info.Protected = m_mad_header.Flags And MAD_FLAG_PROTECTION
    info.FreeFormat = m_mad_header.Flags And MAD_FLAG_FREEFORMAT
    info.Padding = m_mad_header.Flags And MAD_FLAG_PADDING

    If (m_mad_header.Flags And MAD_FLAG_LSF_EXT) Then
        info.MPEGVersion = 2
    ElseIf (m_mad_header.Flags And MAD_FLAG_MPEG_2_5_EXT) Then
        info.MPEGVersion = 2.5
    Else
        info.MPEGVersion = 1
    End If

    info.SamplesPerFrame = GetSamplesPerFrame(info.Layer, info.MPEGVersion)

    ' Framegre in Bytes (nur gltig fr CBR, da abhngig von Bitrate)
    If info.Layer = 1 Then
        info.FrameSize = ((12 * info.CBRBitrate / info.samplerate) + Abs(info.Padding)) * 4
    Else
        If info.Layer = 3 And (m_mad_header.Flags And MAD_FLAG_LSF_EXT) Then
            info.FrameSize = Fix(72 * info.CBRBitrate / info.samplerate) + Abs(info.Padding)
        Else
            info.FrameSize = Fix(144 * info.CBRBitrate / info.samplerate) + Abs(info.Padding)
        End If
    End If
    
    If info.InfoHeaderValid Then
        info.Frames = info.InfoHeader.Frames
    Else
        If info.VBR = VBRXing And (info.VBRXing.Flags And XING_Frames) Then
            info.Frames = info.VBRXing.Frames
        ElseIf info.VBR = VBRVBRI Then
            info.Frames = info.VBRVBRI.VbrIPluginStreamFrames
        ElseIf info.VBR = VBRNone Then
            info.Frames = info.bytes / info.FrameSize
        End If
    End If
    
    info.VBRBitrate = info.bytes * 8 / IPluginStream_Duration * 1000
    
    FindFirstFrame = True
End Function

Private Function GetSamplesPerFrame( _
    ByVal mpeglayer As Integer, _
    ByVal mpegver As Single _
) As Long

    Select Case mpegver
        Case 1
            Select Case mpeglayer
                Case 1:    GetSamplesPerFrame = 384
                Case 2, 3: GetSamplesPerFrame = 1152
            End Select
        Case 2, 2.5
            Select Case mpeglayer
                Case 1:    GetSamplesPerFrame = 384
                Case 2:    GetSamplesPerFrame = 1152
                Case 3:    GetSamplesPerFrame = 576
            End Select
    End Select
End Function

Private Function IPluginStream_StreamRead( _
    sngBuffer() As Single, _
    ByVal lngOffset As Long, _
    ByVal lngSamples As Long, _
    lngSamplesRead As Long _
) As STREAM_RESULT
    
    Dim lngNeeded As Long
    
    IPluginStream_StreamRead = RESULT_SUCCESS
    lngSamplesRead = 0
    
    If m_lngSampleCount = 0 Then
        If m_blnEOF Then
            IPluginStream_StreamRead = RESULT_END_OF_STREAM
        Else
            IPluginStream_StreamRead = RESULT_READ_ERROR
        End If
        Exit Function
    End If
    
    Do
        lngNeeded = Minimum(lngSamples, m_lngSampleCount - m_lngSamplesUsed)
        CpyMem sngBuffer(lngOffset + lngSamplesRead), m_sngSamples(m_lngSamplesUsed), 4 * lngNeeded
        lngSamples = lngSamples - lngNeeded
        m_lngSamplesUsed = m_lngSamplesUsed + lngNeeded
        lngSamplesRead = lngSamplesRead + lngNeeded
        
        If m_lngSamplesUsed = m_lngSampleCount Then
            Select Case FillBuffer()
                Case RESULT_SUCCESS
                    ' nichts tun
                Case RESULT_END_OF_STREAM
                    IPluginStream_StreamRead = RESULT_END_OF_STREAM
                Case Else
                    IPluginStream_StreamRead = RESULT_READ_ERROR
            End Select
        End If
    Loop While (lngSamples > 0) And (Not m_blnEOF)
End Function

Private Function FillBuffer() As STREAM_RESULT
    Dim intBuffer() As Integer
    Dim i As Long
    
    FillBuffer = GetFrame(intBuffer, m_lngSampleCount)
    
    For i = 0 To m_lngSampleCount - 1
        m_sngSamples(i) = intBuffer(i) / 32768
    Next
    m_lngSamplesUsed = 0
End Function

Private Function IPluginStream_StreamSeek(ByVal ms As Long) As STREAM_RESULT
    Dim lngSeekOffset As Long
    Dim btFrame() As Byte
    
    If (ms > IPluginStream_Duration) Or (ms < 0) Then
        IPluginStream_StreamSeek = RESULT_INVALID_PARAM
        Exit Function
    End If
    
    If ms = 0 Then
        mad_timer_set m_mad_timer, 0, ms, 1000
        mad_frame_mute m_mad_frame
        mad_synth_mute m_mad_synth
        
        FileSeek m_hFile, m_lngFirstFrame, FILE_BEGIN
        m_blnEOF = False
    Else
        With m_udtInfo
            If (.VBR = VBRXing) And (.VBRXing.Flags And XING_TOC) Then
                lngSeekOffset = SeekXing(.VBRXing.Toc, _
                                         .bytes, _
                                         ms / IPluginStream_Duration * 100)
                                         
            ElseIf .VBR = VBRVBRI Then
                lngSeekOffset = SeekVBRI(.SamplesPerFrame, _
                                         .VBRVBRI, _
                                         .samplerate, _
                                         ms)
                                         
            ElseIf .VBR = VBRNone Then
                lngSeekOffset = .bytes * (ms / IPluginStream_Duration)
                
            End If
        End With
        
        mad_timer_set m_mad_timer, 0, ms, 1000
        mad_frame_mute m_mad_frame
        mad_synth_mute m_mad_synth
        
        FileSeek m_hFile, lngSeekOffset + m_lngFirstFrame, FILE_BEGIN
        m_blnEOF = False
    End If
    
    IPluginStream_StreamSeek = RESULT_SUCCESS
End Function

' nchsten gltigen Frame finden
Private Function ReSync() As Boolean
    Dim lngRead         As Long
    Dim lngReadAbs      As Long
    Dim lngBufSize      As Long
    Dim lngConsumed     As Long
    Dim btBuffer(4095)  As Byte
    Dim blnValid        As Boolean

    lngReadAbs = FilePosition(m_hFile)
    lngRead = FileRead(m_hFile, VarPtr(btBuffer(0)), UBound(btBuffer) + 1)
    lngBufSize = lngRead
    mad_stream_buffer m_mad_stream, btBuffer(0), lngBufSize
    
    Do
        If mad_header_decode(m_mad_header, m_mad_stream) <> 0 Then
            If m_mad_stream.nError <> MAD_ERROR_NONE Then
                If Not MAD_RECOVERABLE(m_mad_stream.nError) Then
                    If m_mad_stream.nError = MAD_ERROR_BUFLEN Then
                        If FileEnd(m_hFile) Then Exit Do
                        
                        lngConsumed = m_mad_stream.bufend _
                                      - m_mad_stream.next_frame _
                                      - MAD_BUFFER_GUARD
                        
                        If lngConsumed > 0 Then
                            CpyMem btBuffer(0), _
                                   ByVal m_mad_stream.next_frame, _
                                   lngConsumed
                                   
                            lngReadAbs = lngReadAbs - MAD_BUFFER_GUARD
                        Else
                            FileSeek m_hFile, FilePosition(m_hFile) _
                                              - lngBufSize _
                                              + (lngBufSize / 2), _
                                     FILE_BEGIN
                                     
                            lngReadAbs = lngReadAbs - (lngBufSize / 2)
                        End If
                        
                        lngBufSize = lngConsumed
                        
                        lngRead = FileRead(m_hFile, _
                                           VarPtr(btBuffer(lngBufSize)), _
                                           (m_mad_stream.bufend _
                                            - m_mad_stream.buffer) _
                                           - lngConsumed)

                        lngBufSize = lngBufSize + lngRead
                        lngReadAbs = lngReadAbs + lngRead
                        
                        If lngRead = 0 Then Exit Do
                        
                        mad_stream_buffer m_mad_stream, _
                                          btBuffer(0), _
                                          lngBufSize
                    Else
                        Exit Do
                    End If
                End If
            End If
        Else
            blnValid = True
            Exit Do
        End If
    Loop
    
    If Not blnValid Then Exit Function
    
    FileSeek m_hFile, lngReadAbs _
                      + (m_mad_stream.this_frame _
                         - m_mad_stream.buffer), _
             FILE_BEGIN
    
    ReSync = True
End Function

' VBR Seek mit VBRI Interpolationstabelle
Private Function SeekVBRI( _
    ByVal SamplesPerFrame As Long, _
    vbrihdr As VBRIHeader, _
    ByVal samplerate As Long, _
    ByVal EntryTimeInMilliSeconds As Long _
) As Long

    Dim i                       As Long
    Dim SeekPoint               As Long
    Dim fraction                As Long
    
    Dim TotalDuration           As Single
    Dim DurationPerVbriFrames   As Single
    Dim AccumulatedTime         As Single
    
    TotalDuration = (vbrihdr.VbrIPluginStreamFrames _
                     * SamplesPerFrame) _
                    / samplerate * 1000

    DurationPerVbriFrames = TotalDuration / (vbrihdr.VbriTableSize + 1)
    
    If EntryTimeInMilliSeconds > TotalDuration Then
        EntryTimeInMilliSeconds = TotalDuration
    End If
    
    Do While (AccumulatedTime <= EntryTimeInMilliSeconds)
        SeekPoint = SeekPoint + vbrihdr.VbriTable(i)
        AccumulatedTime = AccumulatedTime + DurationPerVbriFrames
        i = i + 1
    Loop
    
    With vbrihdr
        fraction = (((AccumulatedTime - EntryTimeInMilliSeconds) / DurationPerVbriFrames) _
                    + (1# / (2# * .VbriEntryFrames))) * .VbriEntryFrames
        SeekPoint = SeekPoint - ((.VbriTable(i - 1) * fraction) / .VbriEntryFrames)
    End With
    
    SeekVBRI = SeekPoint
End Function

' VBR Seek mit Xing Interpolationstabelle
Private Function SeekXing( _
    Toc() As Byte, _
    ByVal file_bytes As Long, _
    ByVal percent As Single _
) As Long

    Dim a   As Long
    Dim fa  As Single
    Dim fb  As Single
    Dim fx  As Single
    
    If percent < 0 Then percent = 0
    If percent > 100 Then percent = 100
    
    a = percent
    If a > 99 Then a = 99
    fa = Toc(a)
    
    If a < 99 Then
        fb = Toc(a + 1)
    Else
        fb = 256
    End If
    
    fx = fa + (fb - fa) * (percent - a)
    
    SeekXing = (1 / 256) * fx * file_bytes
End Function

Private Function FindID3v1( _
    ByRef Position As Long, _
    ByRef Size As Long _
) As Boolean

    Dim lngOldFPos      As Long
    Dim btBuffer(127)   As Byte
    
    lngOldFPos = FilePosition(m_hFile)
    
    FileSeek m_hFile, FileLength(m_hFile) - 128, FILE_BEGIN
    FileRead m_hFile, VarPtr(btBuffer(0)), UBound(btBuffer) + 1
    
    If BytesToAnsi(btBuffer, 0, 3) = "TAG" Then
        Position = FileLength(m_hFile) - 128
        Size = 128
        FindID3v1 = True
    End If
    
    FileSeek m_hFile, lngOldFPos, FILE_BEGIN
End Function

' ID3v2 Gre ermitteln und unbedingt berspringen,
' eventuell Pseudoframes darin enthalten
Private Function FindID3v2( _
    ByRef Position As Long, _
    ByRef Size As Long _
) As Boolean

    Dim btBuffer(2& * 2048& - 1&) As Byte
    Dim lngOldFPos      As Long
    Dim i               As Long
    Dim udeID3flags     As ID3v2Flags
    Dim btID3size(3)    As Byte
    
    lngOldFPos = FilePosition(m_hFile)
    
    FileSeek m_hFile, 0, FILE_BEGIN
    FileRead m_hFile, VarPtr(btBuffer(0)), UBound(btBuffer) + 1
    
    'For i = 0 To UBound(btBuffer) - 3
        If BytesToAnsi(btBuffer, 0, 3) = "ID3" Then
            Position = i
            
            On Error GoTo ErrorHandler
            Size = GetTagSize(m_hFile.handle, i)
            On Error GoTo 0
            
            FindID3v2 = True
            
            'Exit For
        End If
    'Next
    
ErrorHandler:
    FileSeek m_hFile, lngOldFPos, FILE_BEGIN
End Function

Private Function BytesToAnsi( _
    buf() As Byte, _
    offset As Long, _
    Length As Long _
) As String

    Dim strOut  As String
    Dim i       As Long
    
    For i = offset To offset + Length - 1
        strOut = strOut & Chr(buf(i))
    Next
    
    BytesToAnsi = strOut
End Function

Public Property Get EndOfFile() As Boolean
    EndOfFile = m_blnEOF
End Property

Private Property Get IPluginStream_Duration() As Long
    With m_udtInfo
        IPluginStream_Duration = .Frames * .SamplesPerFrame / .samplerate * 1000
    End With
End Property

Private Property Get IPluginStream_Position() As Long
    IPluginStream_Position = mad_timer_count(m_mad_timer, MAD_UNITS_MILLISECONDS)
End Property

Private Property Get IPluginStream_Samplerate() As Long
    IPluginStream_Samplerate = m_udtInfo.samplerate
End Property

Private Property Get IPluginStream_Channels() As Long
    IPluginStream_Channels = m_udtInfo.Channels
End Property

Private Sub Class_Initialize()
    m_hFile.handle = INVALID_HANDLE
    Set m_clsInfo = New StreamTags
    ReDim m_sngSamples(MADBufferSize) As Single
End Sub


'''''''''''''''''''''''''''''''''''
' Mad Helferlein
'''''''''''''''''''''''''''''''''''

Private Function MAD_RECOVERABLE(ByVal nError As Long) As Boolean
    MAD_RECOVERABLE = (nError And &HFF00)
End Function

Private Function MAD_NCHANNELS(hdr As mad_header) As Long
    If hdr.mode = MAD_MODE_SINGLE_CHANNEL Then
        MAD_NCHANNELS = 1
    Else
        MAD_NCHANNELS = 2
    End If
End Function

Private Function MAD_NSBSAMPLES(hdr As mad_header) As Long
    If hdr.Layer = MAD_LAYER_I Then
        MAD_NSBSAMPLES = 12
    Else
        If hdr.Layer = MAD_LAYER_III And (hdr.Flags And MAD_FLAG_LSF_EXT) Then
            MAD_NSBSAMPLES = 18
        Else
            MAD_NSBSAMPLES = 36
        End If
    End If
End Function

Private Function mad_timer_count( _
    tmr As mad_timer_t, _
    ByVal units As mad_units _
) As Long

    mad_timer_count = mad_timer_count_api(tmr.seconds, tmr.fraction, units)
End Function

Private Sub mad_timer_add( _
    tmr1 As mad_timer_t, _
    tmr2 As mad_timer_t _
)

    mad_timer_add_api tmr1, tmr2.seconds, tmr2.fraction
End Sub

Private Sub mad_header_finish(header As mad_header)
    ' der Grndlichkeit halber
End Sub

Private Sub mad_synth_finish(synth As mad_synth)
    ' der Grndlichkeit halber
End Sub

Private Sub mad_timer_reset(tmr As mad_timer_t)
    tmr.fraction = 0
    tmr.seconds = 0
End Sub

''''''''''''''''''''''''''''''
' API File Handling
''''''''''''''''''''''''''''''

Private Function IsValidFile(ByVal strFile As String) As Boolean
    Dim hInp    As hFile

    hInp = FileOpen(strFile, GENERIC_READ, FILE_SHARE_READ)
    IsValidFile = hInp.handle <> INVALID_HANDLE
    FileClose hInp
End Function

Private Function FileOpen( _
    ByVal strFile As String, _
    Optional access As FILE_ACCESS_RIGHTS = GENERIC_READ Or GENERIC_WRITE, _
    Optional share As FILE_SHARE_RIGHTS = FILE_SHARE_READ Or FILE_SHARE_WRITE, _
    Optional method As FILE_OPEN_METHOD = OPEN_EXISTING _
) As hFile

    FileOpen.handle = CreateFile(strFile, access, share, ByVal 0&, method, 0, 0)
    FileOpen.path = strFile
End Function

Private Sub FileClose(filehandle As hFile)
    CloseHandle filehandle.handle
    filehandle.handle = INVALID_HANDLE
End Sub

Private Function FileRead( _
    filehandle As hFile, _
    ByVal ptr As Long, _
    ByVal bytes As Long _
) As Long

    Dim dwRead  As Long
    Dim lngRet  As Long

    If filehandle.handle = INVALID_HANDLE Then Exit Function

    lngRet = ReadFile(filehandle.handle, ByVal ptr, bytes, dwRead, 0&)

    If lngRet = 1 Then
        FileRead = dwRead
    Else
        FileRead = -1
    End If
End Function

Private Function FileWrite( _
    filehandle As hFile, _
    ByVal ptr As Long, _
    ByVal bytes As Long _
) As Long

    Dim dwWritten   As Long
    Dim lngRet      As Long

    If filehandle.handle = INVALID_HANDLE Then Exit Function

    lngRet = WriteFile(filehandle.handle, ByVal ptr, bytes, dwWritten, 0&)

    If lngRet = 1 Then
        FileWrite = dwWritten
    Else
        FileWrite = -1
    End If
End Function

Private Function FileSeek( _
    filehandle As hFile, _
    ByVal bytes As Long, _
    ByVal method As SEEK_METHOD _
) As Long

    FileSeek = SetFilePointer(filehandle.handle, bytes, 0, method)
End Function

Private Function FilePosition(filehandle As hFile) As Long
    FilePosition = FileSeek(filehandle, 0, FILE_CURRENT)
End Function

Private Function FileLength(filehandle As hFile) As Long
    FileLength = GetFileSize(filehandle.handle, 0)
End Function

Private Function FileEnd(filehandle As hFile) As Boolean
    FileEnd = FilePosition(filehandle) >= FileLength(filehandle)
End Function

' ######################################################
' ######################################################

Private Property Get IPlugin_Description() As String
    IPlugin_Description = "Decodiert MP3 Audio Dateien"
End Property

Private Property Get IPlugin_Name() As String
    IPlugin_Name = "MP3 Stream"
End Property

Private Property Get IPlugin_PluginType() As MonoPluginLib.PLUGIN_TYPE
    IPlugin_PluginType = PLUGIN_STREAM
End Property
